/*
For information on this software or to report bugs please visit http://www.ranton.org
or email ranton.org

Copyright (c) 2002, Richard Neal Anton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

*	Redistributions of source code must retain the above copyright notice, this list
	of conditions and the following disclaimer. 

*	Redistributions in binary form must reproduce the above copyright notice, this list
	of conditions and the following disclaimer in the documentation and/or other materials
	provided with the distribution. 

*	The name of the copyright holder may not be used to endorse or promote products 
	derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdlib.h>
#include <errno.h>
#include <unistd.h>

#include "buffer.h"
#include "util.h"

#undef BUFFERDEBUG

// fragment and buffering stuff for the audio
struct timeval gStartTime; // time the server started at.
long long gCurrFrag = 0;

int gFragSize = 0; // size of a sound fragment.
int gNumFrags = 0; // number of fragments we keep for buffering

// list of fragments buffered
SoundFrag *gFragHead = NULL, *gFragTail = NULL; 
SoundFrag *gFragFree = NULL; // list of free fragments

// this is the sound format we are operating in.
SoundFormat gSoundFormat;

char *gSoundDevName = DEFAULT_DEV; // default sound device
int gSoundDev = -1; // the sound device's fd

// this allocates and sets up the audio buffering stuff.
int setup_audio_buffers(void)
{
	int err = 0;
	int bandwidth, fragsPerSec, numFrags;
	int count;
	SoundFrag *frag = NULL;

	// determine the # of fragments per second.
	bandwidth = GetSoundBandwidth( &gSoundFormat );
	fragsPerSec = bandwidth / gFragSize;
	numFrags = kBufferSeconds * fragsPerSec;
	
	// now allocate them all and put them in the free list.
	for( count = 0; count < numFrags; count++ )
	{
		frag = (SoundFrag *)malloc( sizeof( SoundFrag ) );
		if( !frag )
		{	
			err = ENOMEM;
			goto cleanup;
		}
		frag->fragMem = (unsigned char *)malloc( sizeof(unsigned char)*gFragSize );
		if( !frag->fragMem )
		{
			err = ENOMEM;
			goto cleanup;
		}
		// put it in the free list.
		frag->last = NULL;
		frag->next = gFragFree;
		if( gFragFree )
			gFragFree->last = frag;
		gFragFree = frag;

		frag = NULL;
	}

cleanup:
	if( err )
		report_error("Error in setup_audio_buffers",err);
	// anything in the free list already will get freed anyway,
	// so just free anything that didn't make it in
	if( frag )
	{
		if( frag->fragMem )
			free( (void *)frag->fragMem );
		free( (void *)frag );
	}

	return err;
}

// this frees a list of sound fragments
void free_frag_list(SoundFrag *head)
{
	SoundFrag *frag,*next;

	frag = head;
	while(frag)
	{
		next = frag->next;
		if( frag->fragMem )
			free( (void *)frag->fragMem );
		free( (void *)frag );
		frag = next;
	}
}

// this frees the audio buffers
void free_audio_buffers(void)
{
	free_frag_list( gFragHead );
	gFragHead = NULL;
	gFragTail = NULL;
	free_frag_list( gFragFree );
	gFragFree = NULL;
}

// this reads a new sound fragment, discarding the oldest
// buffered one if necessary to make room, and puts it on
// the tail of the gFragHead/gFragTail list
int read_fragment(void)
{
	int ret, err = 0;
	SoundFrag *frag; 

	if( gFragFree )
	{
		// take it from the free list.
		frag = gFragFree;
		gFragFree = frag->next;
		if( gFragFree )
			gFragFree->last = NULL;
	}
	else
	{
		// ditch the oldest buffered one.
		frag = gFragHead;
		gFragHead = frag->next;
		if( gFragHead )
		{
			gFragHead->last = NULL;
			if( !gFragHead->next )
				gFragTail = gFragHead;
		}
		else
			gFragTail = NULL;
	}
	frag->last = frag->next = NULL;
	// go ahead and put this on the tail of the list, so
	// that if we have an error it will get freed.
	frag->fragNum = -1; // so it doesn't get sent until
		// we have real data in it.
	frag->last = gFragTail;
	if( gFragTail )
		gFragTail->next = frag;
	else
		gFragHead = frag;
	gFragTail = frag;

	// now read a fragment, there should _always_
	// be an entire fragment available since that is
	// how the device allegedly works, and we should
	// only be called if select says there's something
	// to read.
	ret = read( gSoundDev, frag->fragMem, gFragSize );
	if( ret == -1 )
	{
		err = errno;
		goto cleanup;
	}
	else if( ret != gFragSize )
	{
		err = EIO; // for lack of something more relavent
		goto cleanup;
	}
	
	frag->fragNum = gCurrFrag++; // give it a real number
	// ok, all done.

cleanup:
	if( err )
		report_error("Error in read_fragment",err);
	return err;
}

// this writes a sound fragment(the oldest one), then takes it out
// of the used list and puts it in the free list
int write_fragment( void )
{
	SoundFrag *frag = gFragHead;
	int ret = 0;

	if( !frag ) return 0;
	
	#ifdef BUFFERDEBUG
	printf("Writing sound fragment to fd %d, of size %d\n",
		gSoundDev, gFragSize);
	#endif

	ret = write( gSoundDev, frag->fragMem, gFragSize );
	if( ret == -1 )
		return errno;
	else if( ret != gFragSize )
		return EIO; // should write it all at once.
	
	// take it out of the used list
	gFragHead = frag->next;
	if( gFragHead )
		gFragHead->last = NULL;
	else
		gFragTail = NULL;
	
	// and put it in the free list.
	frag->next = gFragFree;
	frag->last = NULL;
	if( gFragFree )
		gFragFree->last = frag;
	gFragFree = frag;


	return 0;
}

