/*
For information on this software or to report bugs please visit http://www.ranton.org
or email ranton.org

Copyright (c) 2002, Richard Neal Anton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

*	Redistributions of source code must retain the above copyright notice, this list
	of conditions and the following disclaimer. 

*	Redistributions in binary form must reproduce the above copyright notice, this list
	of conditions and the following disclaimer in the documentation and/or other materials
	provided with the distribution. 

*	The name of the copyright holder may not be used to endorse or promote products 
	derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <assert.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "util.h"
#include "net.h"
#include "soundio.h"
#include "buffer.h"

#undef CLIENTDEBUG

int gPortNo = kDefaultPort;
int gSocket = -1;

// this holds the recv'ed data until
// there is one whole fragment of it.
unsigned char *gFragBuf = NULL;
int gFragBufPos = 0;



int make_conn(char *hostname);
int start_audio(void);
// recv a fragment from the server.
int recv_fragment( void );

int main(int argc, char **argv)
{
	int ret, err = 0;
	fd_set readfds, writefds;
	int highest_fd;

	if( argc != 2 )
		goto usage;

	#ifdef CLIENTDEBUG
	printf("Making connection\n");
	#endif
	err = make_conn(argv[1]);
	if( err )
		goto cleanup;
	#ifdef CLIENTDEBUG
	printf("Starting audio\n");
	#endif
	err = start_audio();
	if( err )
		goto cleanup;
	
	while( 1 )
	{
		FD_ZERO( &readfds );
		FD_ZERO( &writefds );

		FD_SET( gSoundDev, &writefds );
		FD_SET( gSocket, &readfds );

		highest_fd = gSoundDev;
		if( gSocket > highest_fd )
			highest_fd = gSocket;
		highest_fd++;
		ret = select( highest_fd, &readfds, &writefds, NULL, NULL );
		// no point in having a time out,because we have nothing
		// to do that doesn't involve those 2 file descriptors.
		if( ret == -1 )
		{
			if( errno != EINTR )
				report_error("Error from select",err);
		}
		else if( ret > 0 )
		{
			if( FD_ISSET( gSocket, &readfds ) )
			{
				// receive a fragment of sound from the server.
				err = recv_fragment();
				if( err )
					report_error("Error from recv_fragment",err);
			}

			if( FD_ISSET( gSoundDev, &writefds ) )
			{
				// output some sound.
				err = write_fragment();
				if( err )
					report_error("Error from write_fragment",err);

			}
		}
	}

	goto cleanup; // skip usage
usage:
	fprintf(stderr,"Usage: %s <hostname>\n",argv[0]);
	return 2;
cleanup:
	if( err )
		report_error("Error: ",err);

	if( gSocket != -1 )
	{
		#ifdef CLIENTDEBUG
		printf("Closing socket\n");
		#endif
		ret = close(gSocket);
		if( !err && ret == -1 )
			err = errno;
	}
	if( gSoundDev != -1 )
	{
		ret = close(gSoundDev);
		if( !err && ret == -1 )
			err = errno;
	}
	// free all of the fragment buffers.
	free_audio_buffers();
	if( gFragBuf )
		free( (void *)gFragBuf );

	if( err )
		report_error( "", err );
	return err ? 1 : 0;
}

int make_conn(char *hostname)
{
	int ret, err = 0;
	int fd = -1;
	struct protoent *proto;
	int protoNum;
	struct sockaddr_in address;
	struct hostent *host;
	
	// get the protocol number
	proto = getprotobyname( "tcp" );
	if( !proto )
	{
		err = errno;
		goto cleanup;
	}
	protoNum = proto->p_proto;

	// create a socket.
	ret = socket( AF_INET, SOCK_STREAM, protoNum );
	if( ret == -1 )
	{
		err = errno;
		goto cleanup;
	}
	gSocket = fd = ret;

	// connect to the server.
	#ifdef CLIENTDEBUG
	printf("Getting server host name\n");
	#endif
	host = gethostbyname( hostname );
	if( !host )
	{
		if( err == NETDB_INTERNAL )
			err = errno;
		else
			err = h_errno; // don't think these conflict, who knows
		goto cleanup;
	}

	// zero it out to start with.
	memset( (void *)&address, 0, sizeof(address) );
	address.sin_family = host->h_addrtype;
	address.sin_port = htons( gPortNo );
	assert( host->h_length == sizeof( address.sin_addr ) );

	memcpy( (void *)&address.sin_addr, (void *)host->h_addr,
		sizeof(address.sin_addr) );
	
	#ifdef CLIENTDEBUG
	printf("Connecting to server\n");
	#endif
	ret = connect( gSocket, (struct sockaddr *)&address, 
		sizeof( address ) );
	if( ret == -1 )
	{
		err = errno;
		goto cleanup;
	}
	
	// set it to non-blocking, now that we're connected;
	// connecting is a pain in the ass on non-blocking sockets.
	ret = fcntl( fd, F_SETFL, O_NONBLOCK );
	if( ret == -1 )
	{
		err = errno;
		goto cleanup;
	}

cleanup:
	
	return err;
}

int start_audio(void)
{
	int ret, err = 0;

	ret = open( gSoundDevName, O_WRONLY );
	if( ret == -1 )
		return errno;
	gSoundDev = ret;

	// put it in the appropriate mode.
	// NOTE: it would be better to get the mode
	// from the server.

	gSoundFormat.format = AFMT_S16_NE;
	gSoundFormat.channels = 2;
	gSoundFormat.speed = 44100;

	err = SetSoundFormat( gSoundDev, &gSoundFormat );
	if( err )
		return err;

	// get the fragment size.
	err = GetFragmentSize( gSoundDev, &gFragSize );
	if( err )
		return err;
	
	#ifdef CLIENTDEBUG
	printf("Fragment size is %d\n",gFragSize);
	#endif

	gFragBuf = (unsigned char *)malloc( sizeof(unsigned char)*gFragSize);
	if( !gFragBuf )
		return err;
	
	// set it to non-blocking, now that we're connected;
	// connecting is a pain in the ass on non-blocking sockets.
	ret = fcntl( gSoundDev, F_SETFL, O_NONBLOCK );
	if( ret == -1 )
		return errno;

	// setup the buffers.
	err = setup_audio_buffers();
	if( err )
		return err;

	return 0;
}

// recv a fragment from the server.
int recv_fragment( void )
{
	int ret;
	SoundFrag *frag;

	ret = recv( gSocket, 
		gFragBuf+gFragBufPos, 
		gFragSize-gFragBufPos, 0 );
	if( ret == -1 )
		return errno;

	gFragBufPos += ret;

	if( gFragBufPos < gFragSize )
		return 0;
	
	
	// reset the position in the intermediate buffer.
	gFragBufPos = 0;

	if( gFragFree )
	{
		// take it from the free list.
		frag = gFragFree;
		gFragFree = frag->next;
		if( gFragFree )
			gFragFree->last = NULL;
	}
	else
	{
		// ditch the oldest buffered one.
		frag = gFragHead;
		gFragHead = frag->next;
		if( gFragHead )
		{
			gFragHead->last = NULL;
			if( !gFragHead->next )
				gFragTail = gFragHead;
		}
		else
			gFragTail = NULL;
	}
	frag->last = frag->next = NULL;

	// go ahead and put this on the tail of the list, so
	// that if we have an error it will get freed.
	frag->fragNum = -1; // so it doesn't get sent until
		// we have real data in it.
	frag->last = gFragTail;
	if( gFragTail )
		gFragTail->next = frag;
	else
		gFragHead = frag;
	gFragTail = frag;
	
	// copy it in from the intermediate buffer.
	memcpy( frag->fragMem, gFragBuf, gFragSize );

	frag->fragNum = gCurrFrag++;

	
	return 0;
}

