/*
For information on this software or to report bugs please visit http://www.ranton.org
or email ranton.org

Copyright (c) 2002, Richard Neal Anton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

*	Redistributions of source code must retain the above copyright notice, this list
	of conditions and the following disclaimer. 

*	Redistributions in binary form must reproduce the above copyright notice, this list
	of conditions and the following disclaimer in the documentation and/or other materials
	provided with the distribution. 

*	The name of the copyright holder may not be used to endorse or promote products 
	derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <signal.h>
#include <assert.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "util.h"
#include "net.h"
#include "soundio.h"
#include "buffer.h"

/******************** preprocessor **********************/
#undef SERVDEBUG


/********************* constants ************************/

#define kBacklogSize 5

/********************** structures ************************/
// state data for a sound connection
struct SoundConn_struct {
	struct sockaddr_in addr; // address of the other end.
	int fd; // the socket.
	long long lastFrag; // last sound fragment sent.	
	int markedForDeath;
	struct SoundConn_struct *last,*next; // last,next connection in list.
};

typedef struct SoundConn_struct SoundConn;


/************************ prototypes **********************/
// returns error on fail, 0 on success
int startup_server(void);

// this returns error status
int startup_audio(void);


// this runs the server
int run_server(void);

// cleanup after the server.
int server_cleanup(void);

// send the relevant fragment to the client.
int send_fragment(SoundConn *conn);

// this accepts an incoming connection, and deals with
// the associated state data.
int accept_conn(void);

// kill all the connections that need to die.
int kill_marked_conns(void);

// install our signal handlers.
int install_sig_handlers(void);

// this is our signal handler 
void handle_signal(int sigNum);

/********************* globals ******************************/
int gPortNo = kDefaultPort; // default port number

// file descriptors
int gServerSocket = -1; // the server socket, that we're listening on.



// these are the open connections
SoundConn *gConnHead = NULL;

/**************************************************************/

// returns error on fail, 0 on success
int startup_server(void)
{
	int ret, err = 0;
	int fd = -1;
	struct protoent *proto;
	int protoNum;
	struct sockaddr_in address;
	struct hostent *host;
	char hostname[MAXHOSTNAMELEN+1];

	// get the protocol number
	proto = getprotobyname( "tcp" );
	if( !proto )
	{
		err = errno;
		goto cleanup;
	}
	protoNum = proto->p_proto;

	// create a socket.
	ret = socket( AF_INET, SOCK_STREAM, protoNum );
	if( ret == -1 )
	{
		err = errno;
		goto cleanup;
	}
	fd = ret;
	
	// set it to non-blocking.
	ret = fcntl( fd, F_SETFL, O_NONBLOCK );
	if( ret == -1 )
	{
		err = errno;
		goto cleanup;
	}

	// ok, we now have a socket, we need a port.

	// to get a port we need a sockaddr_in structure.
	ret = gethostname(hostname,MAXHOSTNAMELEN);
	if( ret == -1 )
	{
		err = errno;
		goto cleanup;
	}
	host = gethostbyname( hostname );
	if( !host )
	{
		if( err == NETDB_INTERNAL )
			err = errno;
		else
			err = h_errno; // don't think these conflict, who knows
		goto cleanup;
	}
	// zero it out to start with.
	memset( (void *)&address, 0, sizeof(address) );
	address.sin_family = host->h_addrtype;
	address.sin_port = htons( gPortNo );

	// and actually call bind.
	ret = bind( fd, (struct sockaddr *)&address, sizeof(address) );
	if( ret == -1 )
	{
		err = errno;
		goto cleanup;
	}

	// now we need to listen for connections.
	ret = listen( fd, kBacklogSize );
	if( ret == -1 )
	{
		err = errno;
		goto cleanup;
	}

	#ifdef SERVDEBUG
	printf("Server started\n");
	#endif
cleanup:
	if( err )
		report_error("Error in startup_server",err);

	if( err && fd != -1 )
		ret = close(fd);
	if( !err )
		gServerSocket = fd;

	return err;
}

// this returns error status
int startup_audio(void)
{
	int ret, err = 0, fd = -1;
	
	fd = open( gSoundDevName, O_RDONLY );
	if( fd == -1 )
		return errno;
	gSoundDev = fd;

// NOTE: these should probably be constants.
	gSoundFormat.format = AFMT_S16_NE;
	gSoundFormat.channels = 2;
	gSoundFormat.speed = 44100;

	// NOTE: should we check that we go the requisite shit,
	// and warn somebody, or just hope for the best?
	err = SetSoundFormat( gSoundDev, &gSoundFormat );
	if( err )
		return err;

	// get the fragment size.
	err = GetFragmentSize( gSoundDev, &gFragSize );
	if( err )
		return err;
	#ifdef SERVDEBUG
	printf("Fragment size is %d\n",gFragSize);
	#endif

	// allocate our buffers.
	err = setup_audio_buffers();
	if( err )
		return err;

	// set it to non-blocking.
	ret = fcntl( gSoundDev, F_SETFL, O_NONBLOCK );
	if( ret == -1 )
		return errno;

	#ifdef SERVDEBUG
	if( !err )
		printf("Audio started\n");
	#endif

	return err;
}

// send the relevant fragment to the client.
int send_fragment(SoundConn *conn)
{
	int ret;
	SoundFrag *frag;
	// determine which fragment to send.
	frag = gFragHead;

	while( frag && frag->fragNum <= conn->lastFrag )
	{
		frag = frag->next;
	}
	if( !frag )
		return 0; // nothing to do yet.
	
	// send the fragment.
	ret = send( conn->fd, frag->fragMem, gFragSize, 0 );
	if( ret == -1 )
		return errno;
	#ifdef SERVDEBUG
	printf("Sent %d bytes\n",ret);
	#endif

	assert( ret == gFragSize );

	// update the state of the connection.
	conn->lastFrag = frag->fragNum;

	return 0;
}

typedef void (*SigHandlerProc)(int);

// install our signal handlers.
int install_sig_handlers(void)
{
	SigHandlerProc old_usr1, old_pipe, old_quit, old_abrt;
	// install our signal handlers.
	old_usr1 = signal( SIGUSR1, handle_signal );
	if( old_usr1 == SIG_ERR )
		return errno;
	// sig pipe's are so fucking worthless.
	old_pipe = signal( SIGPIPE, SIG_IGN );
	if( old_pipe == SIG_ERR )
		return errno;
	old_quit = signal( SIGQUIT, handle_signal );
	if( old_quit == SIG_ERR )
		return errno;
	old_abrt = signal( SIGABRT, handle_signal );
	if( old_abrt == SIG_ERR )
		return errno;

	return 0;
}

// this runs the server
int run_server(void)
{
	int ret, err = 0;
	int highest_fd;
	fd_set readfds, writefds, exceptfds;
	struct timeval tv;
	SoundConn *conn;

	err = startup_server(); // create the socket, bind it, and listen
	if( err )
		goto cleanup;

	err = startup_audio();
	if( err )
		goto cleanup;

	// now loop accepting connections,
	// and sending data to any open connections.
	// now, to do this with the sound shit,
	// we need to not block doing anything basically.

	// the man page says we can either set it to non-blocking
	// or we can select the listening socket for reading
	// to check for incoming connection attempts.

	// it's also possible to get open() or fcntl() to 
	// set it up so that we get signals whenever
	// read or write is possible or out of band data
	// has arrived.

	#ifdef SERVDEBUG
	printf("Accepting connections\n");
	#endif

	// record the time the server started.
	ret = gettimeofday( &gStartTime, NULL );
	if( ret == -1 )
	{
		err = errno;
		goto cleanup;
	}

	// install our signal handlers.
	err = install_sig_handlers();
	if( err )
		return err;

	while( 1 )
	{
		// see if anything has happened.
		
		FD_ZERO(&readfds);
		FD_ZERO(&writefds);
		FD_ZERO(&exceptfds);
		highest_fd = 0;
		
		FD_SET( gServerSocket, &readfds );
		if( highest_fd < gServerSocket )
			highest_fd = gServerSocket;

		// check for connections that we can send shit to.
		conn = gConnHead;
		while( conn )
		{
			FD_SET( conn->fd, &writefds );
			if( highest_fd < conn->fd )
				highest_fd = conn->fd;
			conn = conn->next;
		}
		// check to see if we can read from the sound device yet.
		FD_SET( gSoundDev, &readfds );
		if( highest_fd < gSoundDev )
			highest_fd = gSoundDev;

		tv.tv_sec = 0;
		tv.tv_usec = 1000; // 1 millisecond
		
		highest_fd++;
		ret = select( highest_fd, &readfds, &writefds, &exceptfds,
			&tv );
		if( ret == -1 )
		{
			if( errno != EINTR ) // just ignore these and keep going
			{
				err = errno;
				goto cleanup; // NOTE: not sure, but don't think
					// anything we can recover can happen here,
					// except possibly EINTR( signal during the syscall )
			}
		}
		else if( ret ) // we have something to do
		{
			// check to see if we need to read a sound fragment.
			if( FD_ISSET( gSoundDev, &readfds ) )
			{
				// read the next sound fragment.
				err = read_fragment();
				if( err )
					report_error("Error reading fragment",err);
			}

			// check for incoming connections
			if( FD_ISSET( gServerSocket, &readfds ) )
			{
				// we (probably?) need to accept a connection.
				err = accept_conn( );
				if( err )
				{
					// just deal with it.
					report_error("Error accepting connection",err);
				}
			}

			// check for other fd's ready for something.
			conn = gConnHead;
			while( conn )
			{
				err = 0;
				if( FD_ISSET( conn->fd, &writefds ) )
				{
					// send the next fragment that this one is expecting
					// to it, or if that one is no longer buffered,
					// then send it the oldest one still buffered.
					err = send_fragment(conn);
					if( err )
						report_error("Error sending fragment",err);
				}

				// for what error values should we kill
				// the connection?
				if( err == EPIPE )
					conn->markedForDeath = 1;
				conn = conn->next;
			}
			// kill all the connections that need to die.
			err = kill_marked_conns();
			if( err )
				report_error("Error from kill_marked_conns",err);

			// put the memory from sound fragments that have been
			// sent to everyone back into the free list.

		}
	}

cleanup:
	ret = server_cleanup();
	if( !err && ret )
		err = ret;
	if( err )
		report_error("Error in run_server",err);
	// cleanup after the server.

	return err;
}

// cleanup after the server.
int server_cleanup(void)
{
	int ret, err = 0;
	// close the socket.
	if( gServerSocket != -1 )
	{
		ret = close(gServerSocket);
		if( !err && ret == -1 )
			err = errno;
	}
	// close the sound device.
	if( gSoundDev != -1 )
	{
		ret = close(gSoundDev);
		if( !err && ret == -1 )
			err = errno;
	}
	// free the sound buffers.
	free_audio_buffers();

	if( err )
		report_error("Error in server_cleanup",err);
	return err;
}

// this accepts an incoming connection, and deals with
// the associated state data.
int accept_conn(void)
{
	int len, ret, fd = -1, err = 0;
	SoundConn *conn = NULL;
	int send_buf_size, recv_buf_size, val_size;

	#ifdef SERVDEBUG
	printf("Accepting connection\n");
	#endif

	conn = (SoundConn *)malloc( sizeof( SoundConn ) );
	if( !conn )
	{
		err = ENOMEM;
		goto cleanup;
	}
	// zero it out.
	memset( (void *)conn, 0, sizeof( conn ) );
	
	len = sizeof( struct sockaddr_in );
	ret = accept( gServerSocket, (struct sockaddr *)&(conn->addr),&len);
	if( ret == -1 )
	{
		err = errno;
		goto cleanup;
	}
	conn->markedForDeath = 0;
	conn->fd = fd = ret;
	conn->lastFrag = gCurrFrag;// new connections get to wait until
		// the next fragment for sanity.

	// determine what it's buffer sizes are.
	val_size = sizeof(int);
	ret = getsockopt( fd, SOL_SOCKET, SO_SNDBUF, &send_buf_size,
		&val_size );
	if( ret == -1 )
	{
		err = errno;
		goto cleanup;
	}
	assert( val_size == sizeof(int) );

	val_size = sizeof(int);
	ret = getsockopt( fd, SOL_SOCKET, SO_RCVBUF, &recv_buf_size,
		&val_size );
	if( ret == -1 )
	{
		err = errno;
		goto cleanup;
	}
	assert( val_size == sizeof(int) );

	#ifdef SERVDEBUG
	printf("Socket send buf size is %d\n",send_buf_size);
	printf("Socket recv buf size is %d\n",recv_buf_size);
	#endif


cleanup:
	if( err )
		report_error("Error in accept_conn",err);

	if( err )
	{
		if( fd != -1 )
			ret = close(fd);
		if( conn )
			free( (void *)conn );
	}
	else
	{
		// put it in the list
		conn->next = gConnHead;
		conn->last = NULL;
		gConnHead = conn;
	}
	return err;
}

// kill all the connections that need to die.
int kill_marked_conns(void)
{
	SoundConn *conn, *next;
	conn = gConnHead;
	while( conn )
	{
		next = conn->next;
		if( conn->markedForDeath )
		{
			if( close(conn->fd) == -1 )
				report_error("Error closing connection", errno);
			if( conn->last )
				conn->last->next = next;
			else
				gConnHead = next;
			if( conn->next )
				conn->next->last = conn->last;
			free( (void *)conn );
		}
		conn = next;
	}
	return 0;
}

// this is our signal handler 
void handle_signal(int sigNum)
{
	struct timeval tv,diff;
	int ret, err;
	// NOTE: we should suspend signals while we are processing them.

	switch( sigNum )
	{
		case SIGUSR1:
			// print out timing info
			ret = gettimeofday(&tv,NULL);
			if( ret == -1 )
				report_error("Error getting timing info",errno);
			else
			{
				diff.tv_sec = tv.tv_sec - gStartTime.tv_sec;
				diff.tv_usec = tv.tv_usec - gStartTime.tv_usec;
				printf("%Ld fragments read in %ld secs and %ld microsecs\n",
					gCurrFrag,
					diff.tv_sec,
					diff.tv_usec);
			}
			break;
		case SIGQUIT:
		case SIGABRT:	
			ret = server_cleanup();
			exit( ret ? 1 : 0 );
			break;
	}
	// reinstall our signal handlers.
	err = install_sig_handlers();
	if( err )
	report_error("Error reinstalling signal handlers",err);
}

int main(int argc, char **argv)
{
	int err;
	
	err = run_server();
	
	if( err )
		report_error("Error:",err);

	printf("Done.\n");

	return (err ? 1 : 0);
}

