/*
For information on this software or to report bugs please visit http://www.ranton.org
or email ranton.org

Copyright (c) 2002, Richard Neal Anton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

*	Redistributions of source code must retain the above copyright notice, this list
	of conditions and the following disclaimer. 

*	Redistributions in binary form must reproduce the above copyright notice, this list
	of conditions and the following disclaimer in the documentation and/or other materials
	provided with the distribution. 

*	The name of the copyright holder may not be used to endorse or promote products 
	derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdio.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>

#ifndef __SOUND_IO__
#include "soundio.h"
#endif

#ifdef SNDTEST
/****************** globals *******************************/
// either of these should work, the ifdef is to default to the
// one more likely to actually have a device file.

char *devName = DEFAULT_DEV;

int main(int argc, char **argv)
{
	int fd = -1;
	int ret,err = 0;
	SoundFormat sndFmt;
	int fragSize;

	// open the device for reading.
	fd = open( devName, O_RDONLY );
	if( fd == -1 )
	{
		err = errno;
		goto cleanup;
	}

	sndFmt.format = AFMT_S16_NE;
	sndFmt.channels = 2;
	sndFmt.speed = 44100;

	err = SetSoundFormat( fd, &sndFmt );
	if( err )
		goto cleanup;

	printf("Sampling format is %d\n",sndFmt.format);
	printf("Number of channels is %d\n",sndFmt.channels);
	printf("Sampling speed is %dHz\n",sndFmt.speed);

	// get the fragment size.
	ret = ioctl(fd,SNDCTL_DSP_GETBLKSIZE, &fragSize);
	if( ret == -1 )
	{	
		err = errno;
		goto cleanup;
	}
	
	printf("Fragment size is %d bytes\n",fragSize);

cleanup:
	// close the file desc if it's open.
	if( fd != -1 )
	{
		ret = close( fd );
		// don't mask over a prior error.
		if( !err && ret == -1 )
			err = errno;
	}

	// tell the user if we had an error.
	if( err )
		fprintf(stderr,"Error %d - %s\n",err,strerror(err));
	return (err ? 1 : 0);
}
#endif // SNDTEST

// this returns the bandwidth used, in bytes per second
// of a sound format
int GetSoundBandwidth( SoundFormat *format )
{
	int bitsPerSample, bandwidth;


	assert( format->format != AFMT_MPEG );

	switch( format->format )
	{
		case AFMT_IMA_ADPCM:
			// this is 4 bit.
			bitsPerSample = 4;
			break;
		case AFMT_MU_LAW:
		case AFMT_A_LAW:
		case AFMT_U8:
		case AFMT_S8:
			// these are 8 bit
			bitsPerSample = 8;
			break;
		case AFMT_S16_LE:
		case AFMT_S16_BE:
		case AFMT_U16_LE:
		case AFMT_U16_BE:
			// these are 16 bit
			bitsPerSample = 16;
			break;
		case AFMT_MPEG:
			// this one is an undefined bit depth
			return 0; // NOTE: any better ideas?
			break;
	}
	bandwidth = bitsPerSample;
	bandwidth *= format->channels;
	bandwidth *= format->speed;
	bandwidth /= 8;
	return bandwidth;
}

// these sets # of channels, sample format, and sampling speed,
// the actual settings it managed to come up with are returned
// in ioSoundFormat, which is also the input parameters.
// inability to get a desired format,# of channles, or speed
// does not cause this to return an error value.
// if it does return an error, the contents of ioSoundFormat
// are undefined and mostly meaningless.
int SetSoundFormat( int inSound_fd, SoundFormat *ioSoundFormat )
{
	int ret;
	int format,channels,speed;

	format = ioSoundFormat->format;
	channels = ioSoundFormat->channels;
	speed = ioSoundFormat->speed;

	// get the current sample format.
	ret = ioctl(inSound_fd, SNDCTL_DSP_SETFMT, &format);
	if( ret == -1 )
		return errno;

	// set it to signed 16 bit if possible.
	ret = ioctl(inSound_fd, SNDCTL_DSP_SETFMT, &format);
	if( ret == -1 )
		return errno;

	// set the number of channels.
	ret = ioctl(inSound_fd, SNDCTL_DSP_CHANNELS, &channels);
	if( ret == -1 )
		return errno;

	// set the sample rate.
	speed = 44100; // stereo quality.
	ret = ioctl( inSound_fd, SNDCTL_DSP_SPEED, &speed);
	if( ret == -1 )
		return errno;

	ioSoundFormat->format = format;
	ioSoundFormat->channels = channels;
	ioSoundFormat->speed = speed;

	return 0;
}

// this gets the fragment size.
int GetFragmentSize( int inSound_fd, int *outFragSize )
{
	int ret;
	ret = ioctl( inSound_fd, SNDCTL_DSP_GETBLKSIZE, outFragSize );
	return (ret == -1) ? errno : 0;
}

// this returns the deviance of one number from the other
// as a percent of the first number.
// NOTE: this doesn't work if either parameter is negative.
int PercentDeviance( int requested, int actual )
{
	int diff;
	diff = requested - actual;
	diff = (diff < 0) ? -diff : diff;
	return (100*diff)/actual;

}
